/**
 * 
 * Webpack Configuration
 * 
 */


const path = require('path')
const fs = require('fs')
const package = require('./package.json')
const modular = require('./modular.json')
const red = '\x1b[31m'
const green = '\x1b[32m'
const reset = '\x1b[0m'

const modularError = (prop, expected_type, modular_prop) => {

  if(typeof prop !== expected_type) {
    console.log(red, `
    -------------------------------------------------------------
    Error in modular.json -> ${modular_prop}
    Expected {${expected_type}} and get {${typeof prop}}
    -------------------------------------------------------------
    `, reset)
  }
  
}


/**
 * 
 * Loaders
 * 
 */
const htmlLoader        =  require('html-loader')
const HtmlWebpackPlugin =  require('html-webpack-plugin')
const ExtractTextPlugin =  require('extract-text-webpack-plugin')
const OfflinePlugin     =  require('offline-plugin')




/**
 * 
 * About app
 * 
 */

const title = package.name || 'Unknown'
const version = package.version || '0.0.0'









/**
 * 
 * Global options
 * 
 */

const usePolyfill = modular.usePolyfill
const entry = usePolyfill ? ['babel-polyfill', modular.entry] : modular.entry 
const openInBrowser = modular.openInBrowser
const port = modular.port || 8080
const https = modular.https

modularError(usePolyfill, 'boolean', 'usePolyfill')
modularError(modular.entry, 'string', 'entry')
modularError(openInBrowser, 'boolean', 'openInBrowser')
modularError(port, 'number', 'port')



const plugins = []







/**
 * 
 * Production
 * 
 */

const prod_path = path.resolve(__dirname, modular.production)
const filename = modular.output

modularError(modular.production, 'string', 'production')
modularError(filename, 'string', 'output')








/**
 * 
 * Markup
 * 
 */

const template = modular.markup.entry
const favicon = modular.markup.favicon
const htmlDist = modular.markup.output
const minimize = modular.markup.minimize
const removeComments = modular.markup.removeComments
const collapseWhitespace =  modular.markup.collapseWhitespace
const sub = modular.markup.sub


modularError(template, 'string', 'markup.entry')
modularError(favicon, 'string', 'markup.favicon')
modularError(htmlDist, 'string', 'markup.output')
modularError(minimize, 'boolean', 'markup.minimize')
modularError(removeComments, 'boolean', 'markup.removeComments')
modularError(collapseWhitespace, 'boolean', 'markup.collapseWhitespace')
modularError(sub, 'object', 'markup.sub')

plugins.push(new HtmlWebpackPlugin({
  title,
  filename: htmlDist,
   minify: {
       collapseWhitespace
   },
  favicon,
  hash: true,
  template
}))


sub.forEach((sub_markup, index) => {

  modularError(sub_markup.entry, 'string', `markup.sub[${index}].entry`)  
  modularError(sub_markup.output, 'string', `markup.sub[${index}].output`)
  
  if(sub_markup.title)
    modularError(sub_markup.title, 'string', `markup.sub[${index}].title`)
  
  if(sub_markup.favicon)
    modularError(sub_markup.favicon, 'string', `markup.sub[${index}].favicon`)
  
  plugins.push(new HtmlWebpackPlugin({
    title: sub_markup.title || title,
    filename: sub_markup.output,
     minify: {
         collapseWhitespace
     },
    favicon: sub_markup.favicon || favicon,
    hash: true,
    template: sub_markup.entry
  }))
})







/**
 * 
 * Cascading style sheets
 * 
 */
const cssDist = modular.styles.output
const styles = modular.styles.use

let css

if(styles === 'sass')
  css = /\.(sass|scss)$/
else if(styles === 'css')
  css = /\.css$/
else {
  console.log(red, `
  -------------------------------------------------------------  
  Error in modular.json -> styles.use
  Expected {string} "sass" or "css" and get {${styles}}
  -------------------------------------------------------------  
  `, reset)
}

const css_loader = ['css-loader?minimize', 'autoprefixer-loader']
if(styles === 'sass') css_loader.push('sass-loader')


modularError(cssDist, 'string', 'styles.output')
  


plugins.push(new ExtractTextPlugin({
  filename: cssDist,
  disable: false,
  allChunks: true
}))









/**
 * 
 * Progressive Web App
 * 
 */

const manifest = modular.progressive.manifest
const main = modular.progressive.main
const optional = modular.progressive.optional
const additional = modular.progressive.additional
const excludes = modular.progressive.excludes
const output = modular.progressive.output
const entrySW = modular.progressive.entry
const activeProgress = modular.progressive.active && true
const manifest_path = modular.progressive.manifestEntry || null

modularError(manifest, 'boolean', 'progressive.manifest')
modularError(manifest_path, 'string', 'progressive.manifestEntry')
modularError(main, 'object', 'progressive.main')
modularError(optional, 'object', 'progressive.optional')
modularError(additional, 'object', 'progressive.additional')
modularError(excludes, 'object', 'progressive.excludes')
modularError(output, 'string', 'progressive.output')
modularError(activeProgress, 'boolean', 'progressive.active')

// Modify manifest.json name, short_name and description if needed from package.json
const manifest_file = require(manifest_path)

if (manifest && manifest_path && (manifest_file.name !== package.name || manifest_file.short_name !== package.name.split(' ')[0] || manifest_file.description !== package.description)) {
  manifest_file.name = package.name
  manifest_file.description = package.description
  manifest_file.short_name = package.name.split(' ')[0]
  const newManifestObject = JSON.stringify(JSON.parse(JSON.stringify(manifest_file)), null, 2)

  fs.writeFile(manifest_path, newManifestObject, err => {
    if (err) throw err
    console.log(green, '+ Manifest file was updated', reset)
  })
}





const ServiceWorker = {
  navigateFallbackURL: '/',
  minify: true,
  events: true,
  output
}

if(typeof entrySW === 'string')
  ServiceWorker.entry = entrySW
else if(entrySW !== null) {
  console.log(red, `
  -------------------------------------------------------------  
  Error in modular.json -> progressive.entry
  Expected {string|null} and get {${typeof entrySW}}
  -------------------------------------------------------------  
  `, reset)
}

if(activeProgress){
  plugins.push(new OfflinePlugin({
    publicPath: '/',
    safeToUseOptionalCaches: true,
    excludes,
    caches: {
      main,
      additional,
      optional
    },
    ServiceWorker,
    AppCache: {

    }
  }))
}



/**
 * 
 * Assets
 * 
 */

const assets = modular.assets
modularError(assets, 'string', 'assets')
const assets_loader = 'file-loader?name=[hash:6].[ext]&outputPath=' + assets + '/&publicPath=/'

/**
 * 
 * Modules
 * 
 */

  const rules = [
    {
      test: css,
      use: ExtractTextPlugin.extract({
          fallback: 'style-loader',
          use: css_loader,
          publicPath: '/'+modular.production
      })
    },
    {
      test: /\.tsx?$/i,
      use: ['awesome-typescript-loader']
    },
    {
      test: /\.(pug|jade)$/i,
      use: ['html-loader', {
        loader: 'pug-html-loader',
        query: {
          data: { package, manifest: manifest_file },
          pretty: true
        }
      }],
    },
    {
      test: /\.html$/i,
      use: [
        {
          loader: 'html-loader?interpolate!'+template,
          options: {
            minimize,
            removeComments,
            collapseWhitespace
          }
        }
      ],

    },
    {
      test: /\.(webm|mp4|avi|mp3|aac|mov)$/i,
      loader: assets_loader
    },
    {
      test: /\.(png|jpe?g|gif|ico|svg)$/i,
      use: [
        assets_loader,
        'image-webpack-loader'
      ]
    },
    {
      test: /\.(ttf|woff|woff2)$/i,
      loader: assets_loader
    }
  ]


  if(manifest){
    rules.push({
      test: /manifest.json$/,
      use: [         
        'file-loader?name=manifest.json',
        'web-app-manifest-loader',
      ]
    })
  }




/**
 * 
 * 
 * Exporting 
 * 
 */





module.exports = {
    entry,
    output: {
        path: prod_path,
        filename
    },
    resolve: {
      extensions: ['.ts', '.tsx', '.js', '.jsx']
    },
    module: {
        rules
    },
    devServer: {
        contentBase: path.join(__dirname, modular.production),
        compress: true,
        stats: "errors-only",
        open: openInBrowser,
        port,
        https
    },
    plugins
}
